import json
from pathlib import Path
import os

from parameterized import parameterized
import unittest

from .. import parse, dumps, UPLCDialect, eval
from ..cost_model import Budget
from ..util import NodeTransformer
from ..transformer import unique_variables
from ..optimizer import pre_evaluation, remove_traces, remove_force_delay

acceptance_test_path = Path("examples/acceptance_tests")


def acceptance_test_dirs():
    res = []
    for dirpath, dirs, files in sorted(os.walk(acceptance_test_path, topdown=True)):
        if dirs:
            # not a leaf directory
            continue
        res.append(dirpath)
    return res


rewriters = [
    # No transformation -> Checks conformance of implementation of the CEK machine
    NodeTransformer,
    # Rewriting variable names for unique names
    unique_variables.UniqueVariableTransformer,
    # Pre-evaluating subterms - here it will always evaluate the whole expression as there are no missing variables
    pre_evaluation.PreEvaluationOptimizer,
    # Trace removal
    remove_traces.TraceRemover,
    # Force Delay Removal
    remove_force_delay.ForceDelayRemover,
]


class AcceptanceTests(unittest.TestCase):
    # check that none of these transformers change semantics
    @parameterized.expand(
        (f"{path}_{rewriter.__name__}", path, rewriter)
        for path in acceptance_test_dirs()
        for rewriter in rewriters
    )
    def test_acceptance_tests(self, _, dirpath, rewriter):
        files = os.listdir(dirpath)
        input_file = next(f for f in files if f.endswith("uplc"))
        input_file_path = os.path.join(dirpath, input_file)
        with open(input_file_path, "r") as fp:
            input = fp.read()
        output_file = next(f for f in files if f.endswith("uplc.expected"))
        output_file_path = os.path.join(dirpath, output_file)
        with open(output_file_path, "r") as fp:
            output = fp.read().strip()
        try:
            input_parsed = parse(input, filename=input_file_path)
        except Exception:
            self.assertEqual(
                "parse error", output, "Parsing program failed unexpectedly"
            )
            return
        try:
            input_parsed = rewriter().visit(input_parsed)
        except unique_variables.FreeVariableError:
            # will raise an evaluation error anyways
            pass
        comp_res = eval(input_parsed)
        res = comp_res.result
        if isinstance(res, Exception):
            self.assertEqual(
                output, "evaluation failure", "Machine failed but should not fail."
            )
            return
        self.assertNotIn(
            output,
            ("parse error", "evaluation failure"),
            "Program parsed and evaluated but should have thrown error",
        )
        output_parsed = parse(output, filename=output_file_path).term
        res_parsed_unique = unique_variables.UniqueVariableTransformer().visit(res)
        output_parsed_unique = unique_variables.UniqueVariableTransformer().visit(
            output_parsed
        )
        res_dumps = dumps(res_parsed_unique, dialect=UPLCDialect.LegacyAiken)
        output_dumps = dumps(output_parsed_unique, dialect=UPLCDialect.LegacyAiken)
        self.assertEqual(output_dumps, res_dumps, "Program evaluated to wrong output")
        cost_file = next(f for f in files if f.endswith("cost"))
        with open(Path(dirpath).joinpath(cost_file)) as f:
            cost_content = f.read()
        if cost_content == "error":
            return
        cost = json.loads(cost_content)
        expected_spent_budget = Budget(cost["cpu"], cost["mem"])
        if rewriter in (
            pre_evaluation.PreEvaluationOptimizer,
            remove_force_delay.ForceDelayRemover,
        ):
            self.assertGreaterEqual(
                expected_spent_budget,
                comp_res.cost,
                "Program cost more after preeval/trace removal rewrite",
            )
        elif rewriter == remove_traces.TraceRemover:
            pass
        else:
            self.assertEqual(
                expected_spent_budget,
                comp_res.cost,
                "Program evaluated with wrong cost.",
            )



import datetime
import unittest

import hypothesis
from hypothesis import strategies as hst
from frozenlist2 import frozenlist
import pyaiken
from parameterized import parameterized

from .. import *
from ..flat_decoder import unzigzag
from ..flat_encoder import zigzag
from ..optimizer import pre_evaluation
from ..tools import unflatten
from ..transformer import unique_variables, debrujin_variables, undebrujin_variables
from ..ast import *
from .. import lexer
from ..util import NodeVisitor


pos_int = hst.integers(min_value=0, max_value=2**64 - 1)


uplc_data_integer = hst.builds(PlutusInteger, hst.integers())
uplc_data_bytestring = hst.builds(PlutusByteString, hst.binary())


def rec_data_strategies(uplc_data):
    uplc_data_list = hst.builds(
        lambda x: PlutusList(frozenlist(x)), hst.lists(uplc_data)
    )
    uplc_data_constr = hst.builds(
        lambda x, y: PlutusConstr(x, frozenlist(y)),
        pos_int,
        hst.lists(uplc_data),
    )
    uplc_data_map = hst.builds(
        PlutusMap,
        hst.dictionaries(
            hst.one_of(
                uplc_data_integer, uplc_data_bytestring
            ),  # TODO technically constr is legal too, but causes hashing error
            uplc_data,
            dict_class=frozendict.frozendict,
        ),
    )
    return hst.one_of(uplc_data_map, uplc_data_list, uplc_data_constr)


uplc_data = hst.recursive(
    hst.one_of(uplc_data_bytestring, uplc_data_integer),
    rec_data_strategies,
    max_leaves=4,
)
uplc_builtin_boolean = hst.builds(BuiltinBool, hst.booleans())
uplc_builtin_integer = hst.builds(BuiltinInteger, hst.integers())
uplc_builtin_bytestring = hst.builds(BuiltinByteString, hst.binary())
# TODO reenable all text as soon as aiken issue for escaped strings in complex data is fixed
uplc_builtin_string = hst.builds(BuiltinString, hst.text())
# uplc_builtin_string = hst.builds(BuiltinString, hst.from_regex(r"\w*", fullmatch=True))
uplc_builtin_unit = hst.just(BuiltinUnit())


def rec_const_strategies(uplc_constant):
    uplc_builtin_pair = hst.builds(BuiltinPair, uplc_constant, uplc_constant)
    uplc_builtin_list = hst.builds(
        lambda x, y: BuiltinList(frozenlist([x] * y), x),
        uplc_constant,
        hst.integers(min_value=0, max_value=10),
    )
    return hst.one_of(uplc_builtin_list, uplc_builtin_pair)


uplc_constant = hst.recursive(
    hst.one_of(
        uplc_builtin_unit,
        uplc_builtin_string,
        uplc_builtin_bytestring,
        uplc_builtin_integer,
        uplc_builtin_boolean,
        uplc_data,
    ),
    rec_const_strategies,
    max_leaves=4,
)
uplc_error = hst.just(Error())
uplc_name = hst.from_regex(r"[a-z_~'][\w~!'#]*", fullmatch=True)
uplc_builtin_fun = hst.builds(BuiltIn, hst.sampled_from(BuiltInFun))
uplc_variable = hst.builds(Variable, uplc_name)


class UnboundVariableVisitor(NodeVisitor):
    def __init__(self):
        self.scope = []
        self.unbound = set()

    def check_bound(self, name: str):
        if name in self.scope:
            return
        self.unbound.add(name)

    def visit_Lambda(self, node: Lambda):
        self.scope.append(node.var_name)
        self.visit(node.term)
        self.scope.pop()

    def visit_Variable(self, node: Variable):
        self.check_bound(node.name)


@hst.composite
def uplc_expr_all_bound(draw, uplc_expr):
    x = draw(uplc_expr)
    unbound_var_visitor = UnboundVariableVisitor()
    unbound_var_visitor.visit(x)
    unbound_vars = list(unbound_var_visitor.unbound)
    vars = draw(hst.permutations(unbound_vars))
    for v in vars:
        x = Lambda(v, x)
    return x


def rec_expr_strategies(uplc_expr):
    uplc_delay = hst.builds(Delay, uplc_expr)
    uplc_force = hst.builds(Force, uplc_expr)
    uplc_apply = hst.builds(Apply, uplc_expr, uplc_expr)
    uplc_lambda = hst.builds(Lambda, uplc_name, uplc_expr)
    uplc_lambda_bound = uplc_expr_all_bound(uplc_expr)
    return hst.one_of(
        uplc_lambda, uplc_delay, uplc_force, uplc_apply, uplc_lambda_bound
    )


uplc_expr = hst.recursive(
    hst.one_of(uplc_error, uplc_constant, uplc_builtin_fun, uplc_variable),
    rec_expr_strategies,
    max_leaves=10,
)


uplc_version = hst.builds(lambda x, y, z: (x, y, z), pos_int, pos_int, pos_int)
# This strategy also produces invalid programs (due to variables not being bound)
uplc_program_any = hst.builds(Program, uplc_version, uplc_expr)


uplc_expr_valid = uplc_expr_all_bound(uplc_expr)
# This strategy only produces valid programs (all variables are bound)
uplc_program_valid = hst.builds(Program, uplc_version, uplc_expr_valid)

uplc_token = hst.one_of(
    *(hst.from_regex(t, fullmatch=True) for t in lexer.TOKENS.values())
)
uplc_token_concat = hst.recursive(
    uplc_token,
    lambda uplc_token_concat: hst.builds(
        lambda x, y: x.join(y),
        hst.from_regex(r"[\n\r\s]+", fullmatch=True),
        hst.lists(uplc_token_concat, min_size=5),
    ),
)

uplc_program = hst.one_of(uplc_program_any, uplc_program_valid)


class HypothesisTests(unittest.TestCase):
    @hypothesis.given(uplc_program, hst.sampled_from(UPLCDialect))
    @hypothesis.settings(max_examples=1000, deadline=None)
    @hypothesis.example(
        Program(version=(0, 0, 0), term=BuiltinByteString(value=b"")),
        UPLCDialect.LegacyAiken,
    )
    @hypothesis.example(
        Program(version=(0, 0, 0), term=BuiltIn(builtin=BuiltInFun.ConstrData)),
        UPLCDialect.LegacyAiken,
    )
    @hypothesis.example(
        Program(version=(0, 0, 0), term=BuiltinInteger(value=0)),
        UPLCDialect.LegacyAiken,
    )
    @hypothesis.example(
        Program(version=(0, 0, 0), term=BuiltinString("\\")),
        UPLCDialect.LegacyAiken,
    )
    @hypothesis.example(
        Program(version=(0, 0, 0), term=BuiltinString('\\"')),
        UPLCDialect.LegacyAiken,
    )
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=BuiltinList([BuiltinString("\\"), BuiltinString("\\")]),
        ),
        UPLCDialect.LegacyAiken,
    )
    def test_dumps_parse_roundtrip(self, p, dialect):
        self.assertEqual(parse(dumps(p, dialect)), p)

    @hypothesis.given(uplc_program)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=10))
    @hypothesis.example(parse("(program 0.0.0 [(lam a (delay a)) (lam c c)])"))
    @hypothesis.example(
        parse("(program 0.0.0 [(lam a (lam b (error))) (lam _ (error))])")
    )
    @hypothesis.example(
        parse("(program 0.0.0 [(force (builtin mkCons)) (lam _ (error))])")
    )
    @hypothesis.example(
        parse("(program 0.0.0 (lam _ [(builtin mkPairData) (lam ' _)]))")
    )
    @hypothesis.example(parse("(program 0.0.0 (lam _ _))"))
    @hypothesis.example(parse("(program 0.0.0 [(lam x0 (lam _ x0)) (con integer 0)])"))
    @hypothesis.example(parse("(program 0.0.0 [(lam _ (delay _)) (con integer 0)])"))
    @hypothesis.example(parse("(program 0.0.0 (lam _ '))"))
    @hypothesis.example(parse("(program 0.0.0 (delay _))"))
    def test_rewrite_no_semantic_change(self, p):
        code = dumps(p)
        try:
            rewrite_p = unique_variables.UniqueVariableTransformer().visit(parse(code))
        except unique_variables.FreeVariableError:
            return
        try:
            res = eval(p)
            res = unique_variables.UniqueVariableTransformer().visit(res.result)
            res = res.dumps()
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            res = e.__class__
        try:
            rewrite_res = eval(rewrite_p)
            rewrite_res = unique_variables.UniqueVariableTransformer().visit(
                rewrite_res.result
            )
            rewrite_res = rewrite_res.dumps()
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        except Exception as e:
            rewrite_res = e.__class__
        self.assertEqual(
            res,
            rewrite_res,
            f"Two programs evaluate to different results even though only renamed in {code}",
        )

    @hypothesis.given(hst.one_of(hst.text(), uplc_token_concat))
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=1))
    def test_raises_syntaxerror(self, p):
        try:
            parse(p)
        except SyntaxError:
            pass
        except Exception as e:
            self.fail(f"Failed with non-syntaxerror {e}")

    @hypothesis.given(uplc_program)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=1))
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=Apply(
                f=Apply(
                    f=BuiltIn(builtin=BuiltInFun.EqualsString),
                    x=BuiltIn(builtin=BuiltInFun.AddInteger),
                ),
                x=Delay(term=Error()),
            ),
        )
    )
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=Lambda(
                var_name="_",
                term=Apply(
                    f=BuiltIn(
                        builtin=BuiltInFun.AddInteger,
                        applied_forces=0,
                        bound_arguments=[],
                    ),
                    x=Lambda(
                        var_name="_", term=Error(), state=frozendict.frozendict({})
                    ),
                ),
                state=frozendict.frozendict({}),
            ),
        )
    )
    def test_preeval_no_semantic_change(self, p):
        code = dumps(p)
        orig_p = parse(code).term
        rewrite_p = pre_evaluation.PreEvaluationOptimizer().visit(p).term
        params = []
        try:
            orig_res = orig_p
            for _ in range(100):
                if isinstance(orig_res, Exception):
                    break
                if isinstance(orig_res, BoundStateLambda) or isinstance(
                    orig_res, ForcedBuiltIn
                ):
                    p = BuiltinUnit()
                    params.append(p)
                    orig_res = Apply(orig_res, p)
                if isinstance(orig_res, BoundStateDelay):
                    orig_res = Force(orig_res)
                orig_res = eval(orig_res).result
            if not isinstance(orig_res, Exception):
                orig_res = unique_variables.UniqueVariableTransformer().visit(orig_res)
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        try:
            rewrite_res = rewrite_p
            for _ in range(100):
                if isinstance(rewrite_res, Exception):
                    break
                if isinstance(rewrite_res, BoundStateLambda) or isinstance(
                    rewrite_res, ForcedBuiltIn
                ):
                    p = params.pop(0)
                    rewrite_res = Apply(rewrite_res, p)
                if isinstance(rewrite_res, BoundStateDelay):
                    rewrite_res = Force(rewrite_res)
                rewrite_res = eval(rewrite_res).result
            if not isinstance(rewrite_res, Exception):
                rewrite_res = unique_variables.UniqueVariableTransformer().visit(
                    rewrite_res
                )
        except unique_variables.FreeVariableError:
            self.fail(f"Free variable error occurred after evaluation in {code}")
        if not isinstance(rewrite_res, Exception):
            if isinstance(orig_res, Exception):
                self.assertIsInstance(
                    orig_res,
                    RuntimeError,
                    "Original code resulted in something different than a runtime error (exceeding budget) and rewritten result is ok",
                )
            self.assertEqual(
                orig_res,
                rewrite_res,
                f"Two programs evaluate to different results after optimization in {code}",
            )
        else:
            self.assertIsInstance(
                orig_res,
                Exception,
                "Rewrite result was exception but orig result is not an exception",
            )

    @hypothesis.given(uplc_program_valid)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=10))
    @hypothesis.example(Program(version=(0, 0, 0), term=PlutusConstr(0, [])))
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=PlutusByteString(
                b"asdjahsdhjddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd"
            ),
        )
    )
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=PlutusInteger(2**64 + 2),
        )
    )
    @hypothesis.example(
        Program(
            version=(0, 0, 0),
            term=BuiltinString("\x00"),
        )
    )
    @hypothesis.example(
        Program((0, 0, 0), BuiltinPair(BuiltinUnit(), BuiltinString("\x00")))
    )
    @hypothesis.example(Program(version=(0, 0, 0), term=BuiltinString(value="\\")))
    @hypothesis.example(Program((0, 0, 0), BuiltinList([BuiltinString("\x00")])))
    @hypothesis.example(
        Program(version=(0, 0, 0), term=PlutusInteger(value=18446744073709551618))
    )
    def test_flat_encode_pyaiken_hypothesis(self, p):
        self.flat_encode_pyaiken_base(p)

    @parameterized.expand((v.name, v) for v in BuiltInFun)
    def test_flat_encode_pyaiken_builtins(self, _, b: BuiltInFun):
        self.flat_encode_pyaiken_base(Program(version=(0, 0, 0), term=BuiltIn(b)))

    def flat_encode_pyaiken_base(self, p):
        flattened = flatten(p)
        unflattened_aiken_string = pyaiken.uplc.unflat(flattened.hex())
        unflattened_aiken = parse(unflattened_aiken_string)

        p_unique = unique_variables.UniqueVariableTransformer().visit(p)
        unflattened_aiken_unique = unique_variables.UniqueVariableTransformer().visit(
            unflattened_aiken
        )
        self.assertEqual(
            p_unique,
            unflattened_aiken_unique,
            "Aiken unable to unflatten encoded flat or decodes to wrong program",
        )

    @hypothesis.given(hst.integers(), hst.booleans())
    def test_zigzag(self, i, b):
        self.assertEqual(i, unzigzag(zigzag(i, b), b)), "Incorrect roundtrip"

    @hypothesis.given(uplc_program_valid)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=10))
    def test_debrujin_undebrujin(self, p: Program):
        p_unique = unique_variables.UniqueVariableTransformer().visit(p)
        debrujin = debrujin_variables.DeBrujinVariableTransformer().visit(p_unique)
        undebrujin = undebrujin_variables.UnDeBrujinVariableTransformer().visit(
            debrujin
        )
        self.assertEqual(p_unique, undebrujin, "incorrect flatten roundtrip")

    @hypothesis.given(uplc_program_valid)
    @hypothesis.settings(max_examples=1000, deadline=datetime.timedelta(seconds=10))
    @hypothesis.example(
        Program(version=(0, 0, 0), term=PlutusMap(value=frozendict.frozendict({})))
    )
    @hypothesis.example(
        Program(version=(0, 0, 0), term=Lambda(var_name="_", term=Variable(name="_")))
    )
    @hypothesis.example(Program(version=(1, 0, 0), term=BuiltinUnit()))
    def test_flat_unflat_roundtrip(self, p: Program):
        p_unique = unique_variables.UniqueVariableTransformer().visit(p)
        self.assertEqual(p_unique, unflatten(flatten(p)), "incorrect flatten roundtrip")

    # TODO test invalid programs being detected with an free variable error

    @hypothesis.given(uplc_data)
    def test_cbor_plutus_data_roundtrip(self, p: PlutusData):
        encoded = plutus_cbor_dumps(p)
        decoded = data_from_cbor(encoded)
        self.assertEqual(p, decoded, "incorrect cbor roundtrip")

    @hypothesis.given(uplc_data)
    def test_json_plutus_data_roundtrip(self, p: PlutusData):
        encoded = p.to_json()
        decoded = data_from_json_dict(encoded)
        self.assertEqual(p, decoded, "incorrect json roundtrip")



from ..util import NodeTransformer
from ..ast import Apply, ForcedBuiltIn, BuiltInFun, Lambda, Delay, Variable

"""
Removes traces from the AST, as traces have only side-effects and no value
"""


class TraceRemover(NodeTransformer):
    def visit_BuiltIn(self, node: Apply):
        return self.visit_ForcedBuiltIn(node)

    def visit_ForcedBuiltIn(self, node: ForcedBuiltIn):
        if node.builtin == BuiltInFun.Trace:
            return Delay(term=Lambda("y", Lambda("x", Variable("x"))))
        return node



"""
DeBrujin Machine to evaluate UPLC AST
"""
import copy

from .ast import *
from .transformer.unique_variables import UniqueVariableTransformer, FreeVariableError
from .cost_model import CekMachineCostModel, BuiltinCostModel, CekOp, Budget

_LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class ComputationResult:
    result: Union[AST, Exception]
    logs: List[str]
    cost: Budget


def budget_cost_of_op_on_model(
    model: Union[BuiltinCostModel, CekMachineCostModel],
    op: Union[BuiltInFun, CekOp],
    *args: int,
):
    if op not in model.cpu or op not in model.memory:
        return Budget(0, 0)
    return Budget(
        cpu=model.cpu[op].cost(*args),
        memory=model.memory[op].cost(*args),
    )


AST_TO_CEK_OP_MAP = {
    Constant: CekOp.Const,
    Program: CekOp.Startup,
    Variable: CekOp.Var,
    BoundStateLambda: CekOp.Lam,
    ForcedBuiltIn: CekOp.Builtin,
    BoundStateDelay: CekOp.Delay,
    Force: CekOp.Force,
    Apply: CekOp.Apply,
}


class Machine:
    def __init__(
        self,
        budget: Budget,
        cek_machine_cost_model: CekMachineCostModel,
        builtin_cost_model: BuiltinCostModel,
        slippage: int = 5,
    ):
        self.budget = budget
        self.unbudgeted_steps = defaultdict(int)
        self.cek_machine_cost_model = cek_machine_cost_model
        self.builtin_cost_model = builtin_cost_model
        self.slippage = slippage

    # Cost methods

    def spend_budget(self, budget: Budget):
        self.remaining_budget -= budget
        if self.remaining_budget.exhausted():
            raise RuntimeError("Exhausted budget")

    def step_and_maybe_spend(self, term: AST):
        step = [op for ast, op in AST_TO_CEK_OP_MAP.items() if isinstance(term, ast)][0]
        self.unbudgeted_steps[step] += 1
        self.unbudgeted_steps[None] += 1
        if self.unbudgeted_steps[None] >= self.slippage:
            self.spend_unbudgeted_steps()

    def spend_unbudgeted_steps(self):
        for cek_op in CekOp:
            self.spend_budget(
                self.unbudgeted_steps[cek_op]
                * budget_cost_of_op_on_model(self.cek_machine_cost_model, cek_op, 0)
            )
        self.unbudgeted_steps = defaultdict(int)

    # Compute methods

    def eval(self, program: Program):
        try:
            program = UniqueVariableTransformer().visit(program)
        except FreeVariableError as e:
            return ComputationResult(
                e,
                [],
                Budget(0, 0),
            )
        self.remaining_budget = copy.copy(self.budget)
        self.logs = []
        stack = [
            Compute(
                NoFrame(),
                frozendict.frozendict(),
                program.term,
            )
        ]
        self.spend_budget(
            budget_cost_of_op_on_model(self.cek_machine_cost_model, CekOp.Startup, 0)
        )

        try:
            while stack:
                step = stack.pop()
                if isinstance(step, Compute):
                    stack.append(self.compute(step.term, step.ctx, step.env))
                elif isinstance(step, Return):
                    stack.append(self.return_compute(step.context, step.value))
                elif isinstance(step, Done):
                    stack.append(step.term)
                    break
            self.spend_unbudgeted_steps()
            res = stack.pop()
        except Exception as e:
            res = e

        return ComputationResult(
            res,
            self.logs,
            self.budget - self.remaining_budget,
        )

    def compute(self, term: AST, context: Context, state: frozendict.frozendict):
        if isinstance(term, Error):
            raise RuntimeError(f"Execution called Error")
        self.step_and_maybe_spend(term)
        if isinstance(term, Constant):
            return Return(context, term)
        elif isinstance(term, BoundStateLambda):
            return Return(
                context,
                BoundStateLambda(term.var_name, term.term, term.state | state),
            )
        elif isinstance(term, BoundStateDelay):
            return Return(context, BoundStateDelay(term.term, term.state | state))
        elif isinstance(term, Force):
            return Compute(
                FrameForce(
                    context,
                ),
                state,
                term.term,
            )
        elif isinstance(term, ForcedBuiltIn):
            return Return(context, term)
        elif isinstance(term, Apply):
            return Compute(
                FrameApplyArg(
                    state,
                    term.x,
                    context,
                ),
                state,
                term.f,
            )
        elif isinstance(term, Variable):
            try:
                return Return(context, state[term.name])
            except KeyError as e:
                _LOGGER.error(
                    f"Access to uninitialized variable {term.name} in {term.dumps()}"
                )
                raise e
        raise NotImplementedError(f"Invalid term to compute: {term}")

    def return_compute(self, context, value):
        if isinstance(context, FrameApplyFun):
            return self.apply_evaluate(context.ctx, context.val, value)
        elif isinstance(context, FrameApplyArg):
            return Compute(
                FrameApplyFun(
                    value,
                    context.ctx,
                ),
                context.env,
                context.term,
            )
        elif isinstance(context, FrameForce):
            return self.force_evaluate(context.ctx, value)
        elif isinstance(context, NoFrame):
            term = value
            return Done(term)
        raise NotImplementedError(f"Invalid context to return compute: {context}")

    def apply_evaluate(self, context, function, argument):
        if isinstance(function, BoundStateLambda):
            return Compute(
                context,
                function.state | {function.var_name: argument},
                function.term,
            )
        if isinstance(function, ForcedBuiltIn):
            eval_fun = BuiltInFunEvalMap[function.builtin]
            needs_forces = BuiltInFunForceMap[function.builtin]
            if function.applied_forces == needs_forces:
                if eval_fun.__code__.co_argcount == len(function.bound_arguments) + 1:
                    arguments = [*function.bound_arguments, argument]
                    cost = budget_cost_of_op_on_model(
                        self.builtin_cost_model,
                        function.builtin,
                        *(arg.ex_mem() for arg in arguments),
                    )
                    self.spend_budget(cost)
                    if function.builtin == BuiltInFun.Trace:
                        # Hack to add this side effect to the machine
                        self.logs.append(arguments[0].value)
                    res = eval_fun(*arguments)
                else:
                    res = ForcedBuiltIn(
                        function.builtin,
                        function.applied_forces,
                        function.bound_arguments + [argument],
                    )
                return Return(context, res)
        raise RuntimeError(f"Tried to apply arguments to improper object: {function}")

    def force_evaluate(self, context, value):
        if isinstance(value, BoundStateDelay):
            return Compute(context, value.state, value.term)
        if isinstance(value, ForcedBuiltIn):
            needs_forces = BuiltInFunForceMap[value.builtin]
            if value.applied_forces < needs_forces:
                res = ForcedBuiltIn(
                    value.builtin, value.applied_forces + 1, value.bound_arguments
                )
                # Theoretically we could check if the builtin requires 0 arguments, but this is not the case for any function
                return Return(context, res)
        raise RuntimeError(f"Forcing an unforceable object: {context}")



from copy import copy

from ..util import NodeTransformer
from ..ast import *


class OutOfBoundsIndex(ValueError):
    pass


class UnDeBrujinVariableTransformer(NodeTransformer):
    """Reverts the debrujin renaming and at the same time renames variables to a unique name"""

    def __init__(self):
        self.scope = []
        self.count = 0

    def get_name(self, index: str) -> str:
        try:
            return self.scope[-int(index)]
        except IndexError:
            raise OutOfBoundsIndex(f"Variable index {index} is too far back")

    def visit_Lambda(self, node: Lambda):
        nc = copy(node)
        nc.var_name = f"v{self.count}"
        self.count += 1
        self.scope.append(nc.var_name)
        nc.term = self.visit(node.term)
        self.scope.pop()
        return nc

    def visit_Variable(self, node: Variable):
        nc = copy(node)
        nc.name = self.get_name(node.name)
        return nc



import cbor2
import rply.errors

from .compiler_config import DEFAULT_CONFIG
from .cost_model import (
    default_budget,
    Budget,
    CekMachineCostModel,
    default_cek_machine_cost_model_plutus_v2,
    BuiltinCostModel,
    default_builtin_cost_model_plutus_v2,
)
from .lexer import strip_comments, Lexer
from .optimizer.pre_evaluation import PreEvaluationOptimizer
from .optimizer.remove_force_delay import ForceDelayRemover
from .optimizer.remove_traces import TraceRemover
from .parser import Parser
from .machine import Machine
from .ast import AST, UPLCDialect, Program, plutus_cbor_dumps, PlutusByteString, Apply
from .flat_encoder import FlatEncodingVisitor
from .flat_decoder import UplcDeserializer
from .transformer.debrujin_variables import DeBrujinVariableTransformer
from .transformer.undebrujin_variables import UnDeBrujinVariableTransformer
from .transformer.unique_variables import UniqueVariableTransformer
from .util import NoOp


def flatten(x: Program) -> bytes:
    """Returns the properly CBOR wrapped program"""
    x_debrujin = DeBrujinVariableTransformer().visit(x)
    flattener = FlatEncodingVisitor()
    flattener.visit(x_debrujin)
    x_flattened = flattener.bit_writer.finalize()
    return cbor2.dumps(x_flattened)


def unflatten(x_cbor: bytes) -> Program:
    """Returns the program from a singly-CBOR wrapped flat encoding"""
    x = cbor2.loads(x_cbor)
    x_bin = "".join(f"{i:08b}" for i in x)
    reader = UplcDeserializer(x_bin)
    x_debrujin = reader.read_program()
    x_uplc = UnDeBrujinVariableTransformer().visit(x_debrujin)
    return x_uplc


def parse(s: str, filename=None):
    s = strip_comments(s)
    l = Lexer().get_lexer()
    p = Parser().get_parser()
    try:
        tks = l.lex(s)
        program = p.parse(tks)
    except rply.errors.LexingError as e:
        source = s.splitlines()[e.source_pos.lineno - 1]
        raise SyntaxError(
            f"Lexing failed, invalid token: {e.message}",
            (filename, e.source_pos.lineno, e.source_pos.colno, source),
        ) from None
    except rply.errors.ParsingError as e:
        source = s.splitlines()[e.source_pos.lineno - 1] if s else ""
        raise SyntaxError(
            f"Parsing failed, invalid production: {e.message}",
            (filename, e.source_pos.lineno, e.source_pos.colno, source),
        ) from None
    return program


def eval(
    u: AST,
    budget: Budget = default_budget(),
    cek_machine_cost_model: CekMachineCostModel = default_cek_machine_cost_model_plutus_v2(),
    builtin_cost_model: BuiltinCostModel = default_builtin_cost_model_plutus_v2(),
):
    m = Machine(budget, cek_machine_cost_model, builtin_cost_model)
    if not isinstance(u, Program):
        u = Program((1, 0, 0), u)
    return m.eval(u)


def dumps(u: AST, dialect=UPLCDialect.Plutus):
    return u.dumps(dialect)


def compile(
    x: Program,
    config=DEFAULT_CONFIG,
) -> Program:
    """
    Returns compiled UPLC code in... UPLC
    This is useful for applying low-level optimizations to the program
    :param x: the program to compile
    """
    prev_dump = None
    new_dump = x.dumps(UPLCDialect.Plutus)
    while prev_dump != new_dump:
        for step in [
            PreEvaluationOptimizer(skip_traces=not config.remove_traces)
            if config.constant_folding
            else NoOp(),
            TraceRemover() if config.remove_traces else NoOp(),
            ForceDelayRemover() if config.remove_force_delay else NoOp(),
        ]:
            x = step.visit(x)
        prev_dump = new_dump
        new_dump = x.dumps(UPLCDialect.Plutus)
    for step in [
        UniqueVariableTransformer() if config.unique_variable_names else NoOp(),
    ]:
        x = step.visit(x)
    return x


def apply(code: Program, *args: AST):
    version = code.version
    code: AST = code.term
    for d in args:
        code: AST = Apply(code, d)
    code = Program(version, code)
    return code



import copy
import dataclasses
import datetime
import enum
import functools
import json
import math
from pathlib import Path
from typing import Dict, Any, Union
from enum import Enum

import pycardano

from .ast import BuiltInFun


class BudgetMode(Enum):
    CPU = "cpu"
    Memory = "memory"


@dataclasses.dataclass
class Budget:
    cpu: int
    memory: int

    def __add__(self, other: "Budget") -> "Budget":
        return Budget(self.cpu + other.cpu, self.memory + other.memory)

    def __sub__(self, other: "Budget") -> "Budget":
        return Budget(self.cpu - other.cpu, self.memory - other.memory)

    def __mul__(self, other: int) -> "Budget":
        return Budget(self.cpu * other, self.memory * other)

    def __isub__(self, other: "Budget") -> "Budget":
        self.cpu -= other.cpu
        self.memory -= other.memory
        return self

    def __iadd__(self, other: "Budget") -> "Budget":
        self.cpu += other.cpu
        self.memory += other.memory
        return self

    def __imul__(self, other: int) -> "Budget":
        self.cpu *= other
        self.memory *= other
        return self

    def __radd__(self, other: "Budget") -> "Budget":
        return Budget(self.cpu + other.cpu, self.memory + other.memory)

    def __rsub__(self, other: "Budget") -> "Budget":
        return Budget(self.cpu - other.cpu, self.memory - other.memory)

    def __rmul__(self, other: int) -> "Budget":
        return Budget(self.cpu * other, self.memory * other)

    def __ge__(self, other: "Budget") -> bool:
        return self.cpu >= other.cpu and self.memory >= other.memory

    def __gt__(self, other: "Budget") -> bool:
        return self.cpu > other.cpu and self.memory > other.memory

    def __lt__(self, other: "Budget") -> bool:
        return self.cpu < other.cpu and self.memory < other.memory

    def __le__(self, other: "Budget") -> bool:
        return self.cpu <= other.cpu and self.memory <= other.memory

    def __eq__(self, other: "Budget") -> bool:
        return self.cpu == other.cpu and self.memory == other.memory

    def exhausted(self):
        return self.cpu < 0 or self.memory < 0


class CostingFun:
    def cost(self, *memories: int) -> int:
        raise NotImplementedError("Abstract cost not implemented")

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        """Parses the arguments of the Plutus Core cost model description"""
        if isinstance(arguments, int):
            return ConstantCost.from_arguments(arguments)
        if "intercept" in arguments and "slope" in arguments and len(arguments) == 2:
            return LinearCost(arguments["intercept"], arguments["slope"])
        raise NotImplementedError("Cost model unknown")

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        """Updates the cost model from the network configuration"""
        raise NotImplementedError("Base model can not update")


@dataclasses.dataclass
class ConstantCost(CostingFun):
    constant: int = 0

    def cost(self, *memories: int) -> int:
        return self.constant

    @classmethod
    def from_arguments(cls, arguments: int):
        return cls(arguments)

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.constant = network_config[f"{prefix}-arguments"]


@dataclasses.dataclass
class LinearCost(CostingFun):
    intercept: int = 0
    slope: int = 0

    def cost(self, *memories: int) -> int:
        return self.intercept + self.slope * memories[0]

    @classmethod
    def from_arguments(cls, arguments: Dict[str, int]):
        return cls(**arguments)

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.intercept = network_config[f"{prefix}-arguments-intercept"]
        self.slope = network_config[f"{prefix}-arguments-slope"]


@dataclasses.dataclass
class Derived(CostingFun):
    model: CostingFun

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        return cls(CostingFun.from_arguments(arguments))

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.model.update_from_network_config(network_config, prefix)


@dataclasses.dataclass
class LinearInX(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(memories[0])


@dataclasses.dataclass
class LinearInY(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(memories[1])


@dataclasses.dataclass
class LinearInZ(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(memories[2])


@dataclasses.dataclass
class AddedSizes(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(sum(memories))


@dataclasses.dataclass
class SubtractedSizes(Derived):
    minimum: int = 0

    def cost(self, *memories: int) -> int:
        return self.model.cost(max(memories[0] - memories[1], self.minimum))

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        min = arguments["minimum"]
        del arguments["minimum"]
        return cls(
            CostingFun.from_arguments(arguments),
            min,
        )

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.model.update_from_network_config(network_config, prefix)
        self.minimum = network_config[f"{prefix}-arguments-minimum"]


@dataclasses.dataclass
class MultipliedSizes(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(math.prod(memories))


@dataclasses.dataclass
class MinSize(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(min(memories))


@dataclasses.dataclass
class MaxSize(Derived):
    def cost(self, *memories: int) -> int:
        return self.model.cost(max(memories))


@dataclasses.dataclass
class LinearOnDiagonal(CostingFun):
    model_on_diagonal: LinearCost
    model_off_diagonal: ConstantCost = dataclasses.field(
        default_factory=lambda: ConstantCost(0)
    )

    def cost(self, x: int, y: int) -> int:
        if x == y:
            return self.model_on_diagonal.cost(x)
        return self.model_off_diagonal.cost(x, y)

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        const_model = ConstantCost(arguments["constant"])
        linear_model = LinearCost(arguments["intercept"], arguments["slope"])
        return cls(linear_model, const_model)

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.model_off_diagonal.constant = network_config[
            f"{prefix}-arguments-constant"
        ]
        self.model_on_diagonal.update_from_network_config(network_config, prefix)


@dataclasses.dataclass
class ConstAboveDiagonal(CostingFun):
    model_below_equal_diagonal: CostingFun
    model_above_diagonal: ConstantCost = dataclasses.field(
        default_factory=lambda: ConstantCost(0)
    )

    def cost(self, x: int, y: int) -> int:
        if x > y:
            return self.model_above_diagonal.cost(x, y)
        return self.model_below_equal_diagonal.cost(x, y)

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        model = parse_costing_fun(arguments["model"])
        const = ConstantCost(arguments["constant"])
        return cls(model, const)

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.model_above_diagonal.constant = network_config[
            f"{prefix}-arguments-constant"
        ]
        self.model_below_equal_diagonal.update_from_network_config(
            network_config, f"{prefix}-arguments-model"
        )


@dataclasses.dataclass
class ConstBelowDiagonal(CostingFun):
    model_above_equal_diagonal: CostingFun
    model_below_diagonal: ConstantCost = dataclasses.field(
        default_factory=lambda: ConstantCost(0)
    )

    def cost(self, x: int, y: int) -> int:
        if x >= y:
            return self.model_above_equal_diagonal.cost(x, y)
        return self.model_below_diagonal.cost(x, y)

    @classmethod
    def from_arguments(cls, arguments: Union[Dict[str, Any]]):
        model = parse_costing_fun(arguments["model"])
        const = ConstantCost(arguments["constant"])
        return cls(model, const)

    def update_from_network_config(
        self, network_config: Dict[str, int], prefix: str = ""
    ):
        self.model_below_diagonal.constant = network_config[
            f"{prefix}-arguments-constant"
        ]
        self.model_above_equal_diagonal.update_from_network_config(
            network_config, f"{prefix}-arguments-model"
        )


# TODO automatically parse from
# https://github.com/input-output-hk/plutus/blob/43ecfc3403cf908c55af57c8461e96e8b131b97c/plutus-core/cost-model/data/builtinCostModel.json
# or similar files


@dataclasses.dataclass
class BuiltinCostModel:
    cpu: Dict[BuiltInFun, CostingFun]
    memory: Dict[BuiltInFun, CostingFun]


class CekOp(enum.Enum):
    Const = enum.auto()
    Var = enum.auto()
    Lam = enum.auto()
    Apply = enum.auto()
    Delay = enum.auto()
    Force = enum.auto()
    Builtin = enum.auto()
    Startup = enum.auto()
    Constr = enum.auto()
    Case = enum.auto()


@dataclasses.dataclass
class CekMachineCostModel:
    cpu: Dict[CekOp, ConstantCost]
    memory: Dict[CekOp, ConstantCost]


COSTING_FUN_DICT = {
    fun.__name__: fun
    for fun in (
        ConstBelowDiagonal,
        ConstantCost,
        ConstAboveDiagonal,
        LinearOnDiagonal,
        LinearCost,
        AddedSizes,
        MultipliedSizes,
        MinSize,
        MaxSize,
        LinearInX,
        LinearInY,
        LinearInZ,
        SubtractedSizes,
    )
}


def parse_costing_fun(model: dict):
    type, arguments = model["type"], model["arguments"]
    CamelCaseType = "".join(x.capitalize() for x in type.split("_"))
    costing_fun = COSTING_FUN_DICT[CamelCaseType]
    return costing_fun.from_arguments(arguments)


def parse_builtin_cost_model(model: dict):
    cost_model = BuiltinCostModel({}, {})
    for fun, d in model.items():
        builtin_fun_name = fun[:1].capitalize() + fun[1:]
        builtin_fun = BuiltInFun.__dict__[builtin_fun_name]
        cost_model.memory[builtin_fun] = parse_costing_fun(d["memory"])
        cost_model.cpu[builtin_fun] = parse_costing_fun(d["cpu"])
    return cost_model


def updated_builtin_cost_model_from_network_config(
    builtin_cost_model: BuiltinCostModel, network_config: Dict[str, int]
):
    builtin_cost_model = copy.deepcopy(builtin_cost_model)
    for mode, cost_fun_dicts in (
        ("cpu", builtin_cost_model.cpu),
        ("memory", builtin_cost_model.memory),
    ):
        for builtin, cost_model in cost_fun_dicts.items():
            builtin_fun_name = builtin.name[:1].lower() + builtin.name[1:]
            prefix = f"{builtin_fun_name}-{mode}"
            cost_model.update_from_network_config(network_config, prefix)
    return builtin_cost_model


class PlutusVersion(enum.Enum):
    PlutusV1 = "PlutusV1"
    PlutusV2 = "PlutusV2"


@functools.lru_cache()
def default_builtin_cost_model_base():
    builtinCostModel = (
        Path(__file__)
        .parent.joinpath("cost_model_files")
        .joinpath("base")
        .joinpath("builtinCostModel.json")
    )
    with open(builtinCostModel) as f:
        d = json.load(f)
    return parse_builtin_cost_model(d)


NETWORK_CONFIG_DIR = Path(__file__).parent.joinpath("cost_model_files")


def load_network_config(config_date: datetime.date):
    """
    Loads the network config from the network config directory that was released last before the given date
    """
    latest_date = None
    latest_dir_name = None
    for dir in NETWORK_CONFIG_DIR.iterdir():
        if not dir.is_dir():
            continue
        try:
            dir_date = datetime.date.fromisoformat(dir.name)
        except ValueError:
            continue
        if dir_date > config_date:
            continue
        if latest_date is None or dir_date > latest_date:
            latest_date = dir_date
            latest_dir_name = dir.name
    network_config_dir = NETWORK_CONFIG_DIR.joinpath(latest_dir_name)
    file = None
    for file in network_config_dir.iterdir():
        if file.suffix == "json":
            break
    if file is None:
        raise ValueError("Latest network config could not be loaded")
    with open(file) as f:
        d = json.load(f)
    return d


@functools.lru_cache(maxsize=1)
def latest_network_config():
    """
    Loads the latest network config from the network config directory that is most recent
    """
    return load_network_config(datetime.date.today())


def latest_network_config_plutus(plutus_version: PlutusVersion):
    return latest_network_config()[plutus_version.value]


def default_builtin_cost_model(plutus_version: PlutusVersion):
    return updated_builtin_cost_model_from_network_config(
        default_builtin_cost_model_base(), latest_network_config_plutus(plutus_version)
    )


@functools.lru_cache()
def default_builtin_cost_model_plutus_v1():
    return default_builtin_cost_model(PlutusVersion.PlutusV1)


@functools.lru_cache()
def default_builtin_cost_model_plutus_v2():
    return default_builtin_cost_model(PlutusVersion.PlutusV2)


def parse_cek_machine_cost_model(model: dict):
    cost_model = CekMachineCostModel({}, {})
    for op, d in model.items():
        enum_name = str(op)[len("cek") :][: -len("Cost")]
        cek_op = CekOp.__dict__[enum_name]
        cost_model.memory[cek_op] = ConstantCost(d["exBudgetMemory"])
        cost_model.cpu[cek_op] = ConstantCost(d["exBudgetCPU"])
    return cost_model


@functools.lru_cache()
def default_cek_machine_cost_model_base():
    builtinCostModel = (
        Path(__file__)
        .parent.joinpath("cost_model_files")
        .joinpath("base")
        .joinpath("cekMachineCosts.json")
    )
    with open(builtinCostModel) as f:
        d = json.load(f)
    return parse_cek_machine_cost_model(d)


def updated_cek_machine_cost_model_from_network_config(
    cek_machine_cost_model: CekMachineCostModel, network_config: Dict[str, int]
):
    cek_machine_cost_model = copy.deepcopy(cek_machine_cost_model)
    for mode, cost_fun_dicts in (
        ("CPU", cek_machine_cost_model.cpu),
        ("Memory", cek_machine_cost_model.memory),
    ):
        for cek_op, cost_model in cost_fun_dicts.items():
            cek_op_name = f"cek{cek_op.name}Cost"
            cost_model.constant = network_config[f"{cek_op_name}-exBudget{mode}"]
    return cek_machine_cost_model


def default_cek_machine_cost_model(plutus_version: PlutusVersion):
    return updated_cek_machine_cost_model_from_network_config(
        default_cek_machine_cost_model_base(),
        latest_network_config_plutus(plutus_version),
    )


@functools.lru_cache()
def default_cek_machine_cost_model_plutus_v1():
    return default_cek_machine_cost_model(PlutusVersion.PlutusV1)


@functools.lru_cache()
def default_cek_machine_cost_model_plutus_v2():
    return default_cek_machine_cost_model(PlutusVersion.PlutusV2)


def default_budget():
    return Budget(memory=14000000, cpu=10000000000)



from ..util import NodeTransformer
from ..ast import *

"""
Removes occurrences of force(delay(x)) and replaces them with x
"""


class ForceDelayRemover(NodeTransformer):
    def visit_Force(self, node: Force):
        if isinstance(node.term, Delay):
            return self.visit(node.term.term)
        return super().generic_visit(node)
